package dev.fyloz.colorrecipesexplorer.utils

import com.fasterxml.jackson.core.JsonParser
import com.fasterxml.jackson.core.TreeNode
import com.fasterxml.jackson.databind.DeserializationContext
import com.fasterxml.jackson.databind.ObjectMapper
import com.fasterxml.jackson.databind.deser.std.UntypedObjectDeserializer
import com.fasterxml.jackson.databind.module.SimpleModule
import com.fasterxml.jackson.module.kotlin.jacksonObjectMapper
import dev.fyloz.colorrecipesexplorer.model.account.User
import dev.fyloz.colorrecipesexplorer.model.account.userOutputDto
import io.jsonwebtoken.Claims
import io.jsonwebtoken.Jws
import io.jsonwebtoken.Jwts
import io.jsonwebtoken.SignatureAlgorithm
import io.jsonwebtoken.io.DeserializationException
import io.jsonwebtoken.io.Deserializer
import io.jsonwebtoken.io.Encoders
import io.jsonwebtoken.io.IOException
import io.jsonwebtoken.jackson.io.JacksonDeserializer
import io.jsonwebtoken.security.Keys
import java.util.*
import javax.crypto.SecretKey


data class Jwt<B>(
    val subject: String,
    val secret: String,
    val duration: Long? = null,
    val signatureAlgorithm: SignatureAlgorithm = SignatureAlgorithm.HS512,
    val body: B? = null
) {
    val token: String by lazy {
        val builder = Jwts.builder()
            .signWith(keyFromSecret(secret))
            .setSubject(subject)
            .claim("payload", body)

        duration?.let {
            val expirationMs = System.currentTimeMillis() + it
            val expirationDate = Date(expirationMs)

            builder.setExpiration(expirationDate)
        }

        builder.compact()
    }
}

enum class ClaimType(val key: String) {
    GROUP_ID("groupId"),
    GROUP_NAME("groupName")
}

data class UserJwtBody(
    val groupId: Long?,
    val groupName: String?
)

/** Build a [Jwt] for the given [User]. */
fun User.buildJwt(secret: String, duration: Long?) =
    Jwt(
        subject = this.id.toString(),
        secret,
        duration,
        body = userOutputDto(this)
    )

//class JacksonDeserializer<T>(
//    val claimTypeMap: Map<String, Class<*>>
//) : Deserializer<T> {
//    val objectMapper: ObjectMapper
//
//    init {
//        objectMapper = jacksonObjectMapper()
//
//        val module = SimpleModule()
//        module.addDeserializer(Any::class.java, MappedTypeDeserializer(Collections.unmodifiableMap(claimTypeMap)))
//        objectMapper.registerModule(module)
//    }
//
//    override fun deserialize(bytes: ByteArray?): T {
//        return try {
//            readValue(bytes)
//        } catch (e: IOException) {
//            val msg =
//                "Unable to deserialize bytes into a " + returnType.getName().toString() + " instance: " + e.getMessage()
//            throw DeserializationException(msg, e)
//        }
//    }
//
//    protected fun readValue(bytes: ByteArray?): T {
//        return objectMapper.readValue(bytes, returnType)
//    }
//}
//
//private class MappedTypeDeserializer(
//    private val claimTypeMap: Map<String, Class<*>>
//) : UntypedObjectDeserializer(null, null) {
//    override fun deserialize(parser: JsonParser, context: DeserializationContext): Any {
//        val name: String = parser.currentName()
//        if (claimTypeMap.containsKey(name)) {
//            val type = claimTypeMap[name]!!
//            return parser.readValueAsTree<TreeNode>().traverse(parser.codec).readValueAs(type)
//        }
//        // otherwise default to super
//        return super.deserialize(parser, context)
//    }
//}

class CustomDeserializer<T>(map: Map<String, Class<*>>) {
    private val objectMapper: ObjectMapper = jacksonObjectMapper()
    private val returnType: Class<T> = Object::class.java as Class<T>
}

/** Parses the given [jwt] string. */
inline fun <reified B> parseJwt(jwt: String, secret: String) =
    with(
        Jwts.parserBuilder()
            .deserializeJsonWith(JacksonDeserializer(mapOf("payload" to B::class.java)))
            .setSigningKey(keyFromSecret(secret))
            .build()
            .parseClaimsJws(jwt)
    ) {
        val jwt = Jwt<B>(this.body.subject, secret)

        val payload = this.body.get("payload", B::class.java)
        jwt
    }

/** Creates a base64 encoded [SecretKey] from the given [secret]. */
fun keyFromSecret(secret: String) =
    with(Encoders.BASE64.encode(secret.toByteArray())) {
        Keys.hmacShaKeyFor(this.toByteArray())
    }

/** Gets the claim with the given [claimType] in a [Jws]. */
private inline fun <reified T> Jws<Claims>.getClaim(claimType: ClaimType) =
    this.body.get(claimType.key, T::class.java)
